SystemOrganization addCategory: #'Comet-Core'!
SystemOrganization addCategory: #'Comet-Examples'!

!Socket methodsFor: '*comet-sending' stamp: 'lr 7/12/2006 14:57'!
pushData: aStringOrByteArray
	"Send aStringOrByteArray to the remote host trough the given socket. Push the data immediately and ignore all kinds of errors. Return true if all the data could be sent successfully, in case of problems return false and destroy the socket."

	| bytesSent bytesToSend count |
	bytesSent := 0.
	bytesToSend := aStringOrByteArray size.
	[ [ bytesSent < bytesToSend ] whileTrue: [
		count := self 
			primSocket: socketHandle
			sendData: aStringOrByteArray
			startIndex: bytesSent + 1
			count: (bytesToSend - bytesSent min: 5000).
		bytesSent := bytesSent + count.
		(count = 0 and: [ bytesSent ~= bytesToSend ])
			ifTrue: [ self destroy. ^ false ] ] ]
		on: Error do: [ self destroy. ^ false ].
	^ true! !

SUPrototype subclass: #CTComet
	instanceVariableNames: 'pusher url'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Comet-Core'!

!CTComet class methodsFor: 'configuration' stamp: 'lr 7/12/2006 12:10'!
headerField
	^ '_c'! !

!CTComet class methodsFor: 'configuration' stamp: 'lr 7/12/2006 12:10'!
terminator
	^ Character value: 0! !

!CTComet methodsFor: 'actions' stamp: 'lr 7/3/2006 13:54'!
connect
	self call: 'connect'! !

!CTComet methodsFor: 'private' stamp: 'lr 5/28/2006 22:00'!
defaultArguments
	^ (Array with: self url) , super defaultArguments! !

!CTComet methodsFor: 'private' stamp: 'lr 5/28/2006 22:07'!
defaultPrototype
	^ 'Comet'! !

!CTComet methodsFor: 'private' stamp: 'lr 7/12/2006 12:10'!
defaultUrl
	^ self canvas context actionUrl
		addParameter: self class headerField
		value: self pusher id! !

!CTComet methodsFor: 'actions' stamp: 'lr 7/3/2006 13:54'!
disconnect
	self call: 'disconnect'! !

!CTComet methodsFor: 'initialization' stamp: 'lr 7/3/2006 13:33'!
initialize
	super initialize.
	self create! !

!CTComet methodsFor: 'options' stamp: 'lr 7/3/2006 13:38'!
interval: aNumber
	self optionAt: 'interval' put: aNumber! !

!CTComet methodsFor: 'accessing' stamp: 'lr 7/3/2006 13:44'!
pusher
	^ pusher! !

!CTComet methodsFor: 'accessing' stamp: 'lr 7/12/2006 16:27'!
pusher: aPusher
	self url: (self canvas context actionUrl
		withParameter: self class headerField 
		value: (self session 
			registerPusher: (pusher := aPusher)))! !

!CTComet methodsFor: 'options' stamp: 'lr 7/3/2006 13:37'!
reconnect: aBoolean
	self optionAt: 'reconnect' put: aBoolean! !

!CTComet methodsFor: 'initialization' stamp: 'lr 7/3/2006 13:32'!
setPusher: aPusher
	pusher := aPusher! !

!CTComet methodsFor: 'options' stamp: 'lr 7/12/2006 12:08'!
terminator: aString
	self optionAt: 'terminator' put: self class terminator! !

!CTComet methodsFor: 'options' stamp: 'lr 7/3/2006 13:38'!
type: anObject
	self optionAt: 'type' put: anObject! !

!CTComet methodsFor: 'accessing' stamp: 'lr 7/3/2006 13:33'!
url
	^ url ifNil: [ url := self defaultUrl ]! !

!CTComet methodsFor: 'accessing' stamp: 'lr 7/3/2006 13:34'!
url: anUrl
	url := anUrl! !

Object subclass: #CTHandler
	instanceVariableNames: 'id pusher socket'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Comet-Core'!

!CTHandler class methodsFor: 'instance creation' stamp: 'lr 7/12/2006 14:23'!
on: aPusher
	^ self basicNew initializeOn: aPusher! !

!CTHandler methodsFor: 'private' stamp: 'lr 7/12/2006 21:38'!
basicConnect: aRequest
	socket := aRequest responseStream socket.
	socket pushData: (String streamContents: [ :stream |
		WAResponse new
			headerAt: 'Expires' put: 'Thu, 11 Jun 1980 12:00:00 GMT';
			headerAt: 'Cache-Control' put: 'no-cache, must-revalidate';
			headerAt: 'Pragma' put: 'no-cache';
			contentType: 'text/plain';
			writeOn: stream.
		stream nextPutAll: (String new: 1024 withAll: Character space) ])! !

!CTHandler methodsFor: 'private' stamp: 'lr 7/12/2006 14:33'!
basicDisconnect
	socket ifNil: [ ^ self ].
	socket destroy.
	socket := nil! !

!CTHandler methodsFor: 'private' stamp: 'lr 7/12/2006 13:56'!
basicTerminate
	thisContext swapSender: nil.
	Processor activeProcess terminate
	! !

!CTHandler methodsFor: 'actions' stamp: 'lr 7/12/2006 16:27'!
connect: aRequest
	self mutex critical: [
		self 
			basicDisconnect; 
			basicConnect: aRequest ].
	self pusher register: self.
	self basicTerminate! !

!CTHandler methodsFor: 'accessing' stamp: 'lr 7/12/2006 11:22'!
id
	^ id! !

!CTHandler methodsFor: 'initialization' stamp: 'lr 7/12/2006 14:24'!
initializeOn: aPusher
	id := WAExternalID new asString.
	pusher := aPusher! !

!CTHandler methodsFor: 'accessing' stamp: 'lr 7/12/2006 15:02'!
mutex
	^ self pusher mutex! !

!CTHandler methodsFor: 'accessing' stamp: 'lr 7/12/2006 14:24'!
pusher
	^ pusher! !

!CTHandler methodsFor: 'actions' stamp: 'lr 7/12/2006 16:27'!
push: aString
	socket ifNil: [ ^ false ].
	^ socket pushData: aString! !

Object subclass: #CTPusher
	instanceVariableNames: 'mutex handlers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Comet-Core'!

CTPusher class
	instanceVariableNames: 'process delay'!

!CTPusher class methodsFor: 'initialization' stamp: 'lr 7/12/2006 11:40'!
initialize
	Smalltalk 
		addToStartUpList: self after: WAKom; 
		addToShutDownList: self.
	self startUp! !

!CTPusher class methodsFor: 'private' stamp: 'lr 7/12/2006 11:36'!
pingProcess
	self allInstancesDo: [ :each | each ping ].
	delay wait! !

!CTPusher class methodsFor: 'initialization' stamp: 'lr 7/12/2006 11:37'!
shutDown
	process terminate! !

!CTPusher class methodsFor: 'initialization' stamp: 'lr 7/12/2006 21:39'!
startUp
	delay := Delay forSeconds: 15.
	process := [ [ self pingProcess ] repeat ]
		forkAt: Processor lowestPriority
		named: 'ping'! !

!CTPusher methodsFor: 'accessing' stamp: 'lr 7/12/2006 11:50'!
handlers
	"Answer the current list of handlers."

	^ handlers! !

!CTPusher methodsFor: 'initialization' stamp: 'lr 7/12/2006 15:01'!
initialize
	super initialize.
	mutex := Semaphore forMutualExclusion.
	handlers := Array new! !

!CTPusher methodsFor: 'accessing' stamp: 'lr 7/12/2006 15:01'!
mutex
	"Answer the mutex in use with the reciever."

	^ mutex! !

!CTPusher methodsFor: 'public' stamp: 'lr 7/12/2006 16:28'!
ping
	"Ping all the associated handlers."

	mutex critical: [
		handlers := handlers
			select: [ :each | each push: ' ' ] ]! !

!CTPusher methodsFor: 'public' stamp: 'lr 7/12/2006 21:29'!
push: aString
	"Push aString to all associated handlers."

	| data |
	data := aString asString , (String with: (Character value: 0)).
	mutex critical: [
		handlers := handlers
			select: [ :each | each push: data ] ]! !

!CTPusher methodsFor: 'public' stamp: 'lr 7/12/2006 15:02'!
register: aHandler
	"Register a new handler into this pusher."

	mutex critical: [ 
		handlers := handlers
			copyWith: aHandler ]! !

!WARenderCanvas methodsFor: '*comet' stamp: 'lr 7/12/2006 12:17'!
comet
	^ CTComet on: self! !

WAScriptLibrary subclass: #CTLibrary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Comet-Core'!

!CTLibrary methodsFor: 'scripts' stamp: 'lr 7/12/2006 21:33'!
comet
	^ '/* Seaside Comet API
 *  (c) 2006 Lukas Renggli (renggli@iam.unibe.ch)
 */
Comet = Class.create();
Comet.prototype = {
	initialize: function (url, options) {
		this.url = url;
		this.options = Object.extend({
			type: Comet.HttpRequest,
			reconnect: true,
			terminator: "\x00",
			interval: 250,
			onAction: this.eval
		}, options || {});
		Object.extend(this, this.options.type);
	},
	eval: function (expression) {
		try { eval(expression) }
		catch (e) { alert(e) }
	}
};

Comet.HttpRequest = {
	connect: function() {
		if (!!this.isConnected()) {
			this.position = 0;
			this.transport = Ajax.getTransport();
			this.transport.open("get", this.url, true);
			this.transport.onreadystatechange = this.onStateChange.bind(this);
			this.transport.send(null);
			this.updater = setInterval(this.update.bind(this), this.options.interval);
		}
	},
	disconnect: function() {
		clearInterval(this.updater); 
		this.transport.onreadystatechange = Prototype.emptyFunction;
		this.transport.abort();
	},
	isConnected: function() {
		return this.transport && Ajax.Request.Events[this.transport.readyState] == "Interactive";
	},
	update: function() {
		if (this.isConnected()) {
			var response = this.transport.responseText;
			if (response) {
				var index = response.lastIndexOf(this.options.terminator);
				if (this.position < index) {
					response
						.substring(this.position, index)
						.split(this.options.terminator)
						.each(this.options.onAction);
					this.position = index + this.options.terminator.length;
				}
			}
		}
	},
	onStateChange: function() {
		if (Ajax.Request.Events[this.transport.readyState] == "Complete") {
			this.disconnect();
			if (this.options.reconnect)
				setTimeout(this.connect.bind(this), this.interval);
		}
	}
};'! !

!WriteStream methodsFor: '*comet' stamp: 'lr 7/3/2006 14:14'!
crlf
	self nextPutAll: String crlf! !

WASession subclass: #CTSession
	instanceVariableNames: 'handlers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Comet-Core'!

!CTSession methodsFor: 'processing' stamp: 'lr 7/12/2006 12:10'!
cometRequest: aRequest
	(handlers
		at: (aRequest fields
			at: CTComet headerField
			ifAbsent: [ ^ nil ])
		ifAbsent: [ ^ nil ])
			connect: aRequest! !

!CTSession methodsFor: 'processing' stamp: 'lr 7/12/2006 12:10'!
incomingRequest: aRequest
	aRequest responseStream
		ifNil: [ self error: 'Streaming-server required.' ].
	^ (aRequest fields includesKey: CTComet headerField)
		ifTrue: [ self cometRequest: aRequest ]
		ifFalse: [ super incomingRequest: aRequest ]! !

!CTSession methodsFor: 'initialization' stamp: 'lr 7/12/2006 11:02'!
initializeWithApplication: anApplication
	super initializeWithApplication: anApplication.
	handlers := Dictionary new! !

!CTSession methodsFor: 'public' stamp: 'lr 7/12/2006 14:05'!
registerPusher: aPusher
	| handler |
	handler := CTHandler on: aPusher.
	handlers at: handler id put: handler.
	^ handler id! !

SUFunctionalTest subclass: #CTChatTest
	instanceVariableNames: 'username message'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Comet-Examples'!

CTChatTest class
	instanceVariableNames: 'pusher'!

!CTChatTest class methodsFor: 'initialization' stamp: 'lr 7/12/2006 15:04'!
initialize
	| application |
	pusher := CTPusher new.
	application := self registerAsApplication: 'chat'.
	application preferenceAt: #sessionClass put: CTSession.
	application libraries add: SULibrary; add: CTLibrary! !

!CTChatTest class methodsFor: 'accessing' stamp: 'lr 7/12/2006 15:04'!
pusher
	^ pusher! !

!CTChatTest methodsFor: 'private' stamp: 'lr 7/12/2006 21:25'!
event: aString
	self pusher push: (SUInsertion bottom
		id: 'messages';
		with: '<div class="message">' , 
			'<span class="username">' , username , '</span>' ,
			'<span class="event">' , aString , '</span>' ,
			'</div>')! !

!CTChatTest methodsFor: 'initialization' stamp: 'lr 7/12/2006 12:07'!
initialize
	super initialize.
	username := 'unknown'! !

!CTChatTest methodsFor: 'private' stamp: 'lr 7/12/2006 21:25'!
message: aString
	self pusher push: (SUInsertion bottom
		id: 'messages';
		with: '<div class="message">' , 
			'<span class="username">' , username , '</span>' ,
			'<span class="message">' , aString , '</span>' ,
			'</div>')! !

!CTChatTest methodsFor: 'accessing' stamp: 'lr 7/3/2006 13:40'!
pusher
	^ self class pusher! !

!CTChatTest methodsFor: 'rendering' stamp: 'lr 7/12/2006 16:59'!
renderContentOn: html
	html textInput
		id: 'username';
		value: username;
		onBlur: (html request triggerFormElement: 'username');
		callback: [ :value | username := value. self event: 'joined' ].
	html div
		id: 'messages';
		style: 'height: 150px; overflow: auto;'.
	html textInput
		id: 'message';
		style: 'width: 100%';
		callback: [ :value | self message: value ];
		onEnter: (html request
			triggerFormElement: 'message';
			onSuccess: (SUStream on: '$("message").value = ""')).
	html script: (html comet
		pusher: self pusher;
		connect)! !

!CTChatTest methodsFor: 'rendering' stamp: 'lr 4/11/2006 23:25'!
style
	^ 'div.message {
	margin-top: 3px;
	margin-bottom: 3px;
	background-color: #dddddd;
}
div.message span.username {
	margin-right: 1em;
	font-weight: bold;
}
div.message span.message {
}
div.message span.event {
	font-style: italic;
}
' 
! !
CTPusher initialize!
CTChatTest initialize!
